/*
 *
 * Utility to list devices
 *
 * Copyright (C) 2003 International Business Machines, Inc.
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this program; if not, write to the Free Software Foundation, Inc.,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "sysdiag.h"

/**
 * indent: called before printing a line, it adds indent to the line up to
 *	level passed in.
 * @level: number of spaces to indent.
 */
void indent(int level)
{
	int i;
	for (i = 0; i <= level; i++)
		fprintf(stdout, " ");
}

/**
 * remove_end_newline: removes newline on the end of an attribute value
 * @value: string to remove newline from
 */
void remove_end_newline(char *value)
{
	char *p = value + (strlen(value) - 1);
	if (p != NULL && *p == '\n')
		*p = '\0';
}

/**
 * show_device_children: prints out device subdirs.
 * @sdir: print this directory's subdirectories/children.
 */
void show_device_children(struct sysfs_directory *sdir, int level)
{
	if (sdir != NULL) {
		indent(level);
		fprintf(stdout, "Children:\n");
		if (sdir->subdirs != NULL) {
			char dirname[SYSFS_NAME_LEN];
			struct sysfs_directory *cur = sdir->subdirs;

			while (cur != NULL) {
				indent(level+4);
				if ((sysfs_get_name_from_path(cur->path,
				    dirname, SYSFS_NAME_LEN)) == 0)
					fprintf(stdout, "%s\n", dirname);
				else
					fprintf(stdout, "%s\n", cur->path);
				cur = cur->next;
			}
		}
		if (sdir->links != NULL) {
			struct sysfs_link *curl = sdir->links;
			while (curl != NULL) {
				indent(level+4);
				fprintf(stdout, "%s\n", curl->name);
				curl = curl->next;
			}
		}
	}
}

/**
 * show_subdirectories: prints all subdirectory names at path
 * @path: sysfs path where subdirs are.
 * returns 0 with success or 1 with error.
 */
int show_subdirectories(char *path, int level)
{
	struct sysfs_directory *dir = NULL, *current = NULL;
	char name[SYSFS_NAME_LEN];

	dir = sysfs_open_directory(path);
	if (dir == NULL) {
		fprintf(stderr, "Error opening sysfs directory at %s\n", path);
		return 1;
	}
	if ((sysfs_read_directory(dir)) != 0) {
		fprintf(stderr, "Error reading sysfs directory at %s\n", path);
		return 1;
	}
	current = dir->subdirs;
	while (current != NULL) {
		if ((sysfs_get_name_from_path(current->path, name,
		    SYSFS_NAME_LEN)) != 0) {
			sysfs_close_directory(dir);
			return 1;
		}
		indent(level);
		fprintf(stdout, "%s\n", name);
		current = current->next;
	}
	sysfs_close_directory(dir);
	return 0;
}

/**
 * isbinaryvalue: checks to see if attribute is binary or not.
 * @attr: attribute to check.
 * returns 1 if binary, 0 if not.
 */
int isbinaryvalue(struct sysfs_attribute *attr)
{
	int i;

	if (attr == NULL || attr->value == NULL)
		return 0;

	for (i = 0; i < attr->len; i++)
		if (!(isascii(attr->value[i])))
			return 1;
	return 0;
}

/**
 * show_attribute_value: prints out single attribute value.
 * @attr: attricute to print.
 */
void show_attribute_value(struct sysfs_attribute *attr, int level)
{
	if (attr == NULL)
		return;

	if (attr->method & SYSFS_METHOD_SHOW) {
		if (isbinaryvalue(attr) != 0) {
			int i;
			fprintf(stdout, "\t: ");
			for (i = 0; i < attr->len; i++) {
				if (!(i %16)) {
					fprintf(stdout, "\n");
					indent(level+8);
				} else if (!(i %8))
					fprintf(stdout, " ");
				fprintf(stdout, "%02x ",
					(unsigned char)attr->value[i]);
			}
			fprintf(stdout, "\n");

		} else if (attr->value != NULL && strlen(attr->value) > 0) {
			remove_end_newline(attr->value);
			fprintf(stdout, "\t:%s\n", attr->value);
		} else
			fprintf(stdout, "\n");
	} else {
		fprintf(stdout, "\t: store method only\n");
	} 
}

/**
 * show_attribute: prints out a single attribute
 * @attr: attribute to print.
 */
void show_attribute(struct sysfs_attribute *attr, int level)
{
	char attrname[SYSFS_NAME_LEN];

	if (attr == NULL) 
		return;

	if ((sysfs_get_name_from_path(attr->path, attrname, SYSFS_NAME_LEN)) 
	    != 0)
		return; 

	if ((DIAG_OPTIONS & DIAG_SHOW_VALUE)) {
		if (!((strncmp(attrname, "resource", 8) == 0)
				|| (strncmp(attrname, "config", 6) == 0)
				|| (strncmp(attrname, "pools", 5) == 0)) 
				|| (DIAG_LEVEL > 1 )) {
			if (!(attr->method & SYSFS_METHOD_SHOW))
				return;
			indent(level);
			fprintf(stdout, "%s", attrname);
			show_attribute_value(attr, level);
		}
	} else 
		fprintf(stdout, "\n");
	
}

/**
 * show_attributes: prints out a directory's attributes.
 * @sdir: print this directory's attributes/files.
 */
void show_attributes(struct sysfs_directory *sdir, int level)
{
	if (sdir != NULL && sdir->attributes != NULL) {
		struct sysfs_attribute *cur = sdir->attributes;

		indent(level);
		fprintf (stdout, "Attributes:\n");
		while (cur != NULL) {
			show_attribute(cur, (level+4));
			cur = cur->next;
		}
	}
}

/**
 * show_driver_attributes: prints out driver attributes .
 * @driver: print this driver's attributes.
 */
void show_driver_attributes(struct sysfs_driver *driver, int level)
{
	if (driver != NULL && driver->directory != NULL) {
		struct sysfs_directory *sdir = driver->directory;

		if (sdir->attributes != NULL) {
			struct sysfs_attribute *cur = sdir->attributes;

			indent(level);
			fprintf (stdout, "%s Attributes:\n", driver->name);
			while (cur != NULL) {
				show_attribute(cur, (level+4));
				cur = cur->next;
			}
		}
	}
}

/**
 * show_driver_devices: prints out devices under driver.
 * @driver: print devices bound to this driver.
 */
void show_driver_devices(struct sysfs_driver *driver, int level)
{
	if (driver != NULL && driver->directory != NULL) {
		struct sysfs_directory *sdir = driver->directory;

		if (sdir->links != NULL) {
			struct sysfs_link *cur = sdir->links;

			indent(level);
			fprintf (stdout, "Devices:\n");
			while (cur != NULL) {
				indent(level+4);
				fprintf (stdout, "%s\n", cur->name);
				cur = cur->next;
			}
		}
	}
}

/**
 * show_driver: prints out driver information.
 * @driver: driver to print.
 */
void show_driver(struct sysfs_driver *driver, int level)
{
	struct sysfs_directory *sdir = NULL;

	if (driver != NULL) {
		indent(level);
		fprintf (stdout, "%s\n", driver->name);
		sdir = driver->directory;
		if (sdir != NULL) {
			if (sdir->links != NULL)
				show_driver_devices(driver, (level+4));
			if (sdir->attributes != NULL && (DIAG_OPTIONS
						& DIAG_SHOW_VALUE))
				show_driver_attributes(driver, (level+4));
		}
	}
}


/**
 * show_device: prints out device information.
 * @device: device to print.
 */
void show_device(struct sysfs_device *device, int level)
{
	if (device != NULL && device->directory != NULL) {
		indent(level);
		fprintf (stdout, "%s %s\n", device->bus_id, device->name);
		if (device->directory->subdirs != NULL
		    || device->directory->links != NULL)
			show_device_children(device->directory, (level+4));
		if (device->directory->attributes != NULL && (DIAG_OPTIONS
					& DIAG_SHOW_VALUE))
			show_attributes(device->directory, (level+4));
		if (device->driver != NULL) {
			indent(level+4);
			fprintf (stdout, "Driver: %s\n",
						device->driver->name);
		}
	}
}

/**
 * show_root_device: prints out sys/devices device information.
 * @device: device to print.
 */
void show_root_device(struct sysfs_device *device, int level)
{
	if (device != NULL && device->directory != NULL) {
		indent(level);
		fprintf (stdout, "%s %s\n", device->bus_id, device->name);
		if (device->directory->attributes != NULL && (DIAG_OPTIONS
					& DIAG_SHOW_VALUE))
			show_attributes(device->directory, (level+4));
	}
}


/**
 * show_device_tree: prints out device tree.
 * @root: root device
 */
void show_device_tree(struct sysfs_device *root, int level)
{
	if (root != NULL) {
		struct sysfs_device *cur = NULL;

		if (DEVICE_NAME[0] == 0 || (strcmp(DEVICE_NAME,
						root->bus_id) == 0)) {
			show_root_device(root, level);
		}
		cur = root->children;
		while (cur != NULL) {
			show_device_tree(cur, (level+6));
			cur = cur->next;
		}
	}
}

/**
 * show_sysfs_root: prints out sysfs root device tree
 * @rootname: device root to print.
 * returns 0 with success and 1 with error.
 */
int show_sysfs_root(char *rootname)
{
	struct sysfs_device *root = NULL;
	char path[SYSFS_PATH_MAX];

	if (rootname == NULL) {
		fprintf(stderr, "rootname is null\n");
		errno = EINVAL;
		return 1;
	}
	if (sysfs_get_mnt_path(path, SYSFS_PATH_MAX) != 0) {
		perror("sysfs_get_mnt_path");
		fprintf(stderr, "Error getting sysfs mount point\n");
		exit(1);
	}

	strcat(path, SYSFS_DEVICES_DIR);
	strcat(path, "/");
	strcat(path, rootname);
	root = sysfs_open_device_tree(path);
	if (root == NULL) {
		fprintf(stderr, "Error opening root device %s\n", rootname);
		return 1;
	}
	fprintf(stdout, "\n-------------------------------------\n");
	fprintf(stdout, "Device Tree: %s\n", rootname);
	fprintf(stdout, "-------------------------------------\n");
	show_device_tree(root, 2);
	sysfs_close_device_tree(root);

	return 0;
}


/**
 *
 */
int list_bus()
{
	struct sysfs_bus *bus = NULL;
	struct sysfs_device *curdev = NULL;
	struct sysfs_driver *curdrv = NULL;

	bus = sysfs_open_bus(BUS_NAME);
	if (bus == NULL) {
		fprintf(stderr, "error opening bus: %s\n",BUS_NAME);
		exit(1);
	}
	fprintf(stdout, "\n-------------------------------------\n");
	fprintf(stdout, "Bus: %s\n", BUS_NAME);
	fprintf(stdout, "-------------------------------------\n");
	if ((bus->devices != NULL ) && (DIAG_OPTIONS & DIAG_DEVICE)) {
			curdev = bus->devices;
		if (DEVICE_NAME[0] == 0)
			fprintf(stdout, "Devices:\n");
		while (curdev != NULL) {
			if (DEVICE_NAME[0] == 0 || (strcmp(DEVICE_NAME,
			    curdev->bus_id) == 0))
				show_device(curdev, 2);
			curdev = curdev->next;
		}
	}
	if (bus->drivers != NULL && (DIAG_OPTIONS & DIAG_DRIVER)) {
		curdrv = bus->drivers;
		fprintf(stdout, "Drivers:\n");
		while (curdrv != NULL) {
			if (DRIVER_NAME[0] == 0 || (strcmp(DRIVER_NAME,
			    curdrv->name) == 0))
				show_driver(curdrv, 2);
			curdrv = curdrv->next;
		}

	}
	sysfs_close_bus(bus);
	return 0;

}


/**
 * show_class_device: prints out class device.
 * @dev: class device to print.
 */
void show_class_device(struct sysfs_class_device *dev, int level)
{
	struct sysfs_directory *cur = NULL;
	char dirname[SYSFS_NAME_LEN];

	if (dev != NULL) {
		indent(level);
		fprintf(stdout, "%s\n", dev->name);
		if (dev->directory != NULL && (DIAG_OPTIONS
					& DIAG_SHOW_VALUE)) { 
			show_attributes(dev->directory, (level+4));
			cur = dev->directory->subdirs;
			while (cur != NULL) {
				if ((sysfs_get_name_from_path(cur->path,
				     dirname, SYSFS_NAME_LEN)) == 0) {
					indent(level+4);
					fprintf(stdout, "%s\n", dirname);
				}
				show_attributes(cur, (level+4));
				cur = cur->next;
			}
		}
		if (dev->sysdevice != NULL && (DIAG_OPTIONS & DIAG_DEVICE))
			show_device(dev->sysdevice, (level+4));
		if (dev->driver != NULL && (DIAG_OPTIONS & DIAG_DRIVER))
			show_driver(dev->driver, (level+4));
	}
}


/**
 *
 *
 */
int list_class()
{
	struct sysfs_class *class = NULL;
	struct sysfs_class_device *cur = NULL;

	if (CLASS_NAME[0] == 0) {
		errno = EINVAL;
		return 1;
	}
	class = sysfs_open_class(CLASS_NAME);
	if (class == NULL) {
		fprintf(stderr, "Error opening class %s\n", CLASS_NAME);
		return 1;
	}
	fprintf(stdout, "\n-------------------------------------\n");
	fprintf(stdout, "Class: %s\n", CLASS_NAME);
	fprintf(stdout, "-------------------------------------\n");
	if (class->devices != NULL) {
		cur = class->devices;
		if (CLASS_NAME[0] == 0)
			fprintf(stdout, "Class Devices:\n");
		while (cur != NULL) {
			if (DEVICE_NAME[0] == 0 || (strcmp(DEVICE_NAME,
			    cur->name) == 0))
				show_class_device(cur, 2);
			cur = cur->next;
		}
	}
	sysfs_close_class(class);
	return 0;
}

/**
 * show_all_bus_device()::
 * this will show all the buses and device on those buses on the system.
 * we open the sysfs/bus dir and for each directory call list_bus function
 * by setting the BUS_NAME variable.
 */
int show_all_device(char *path)
{
	struct sysfs_directory *dir = NULL, *current = NULL;
	char name[SYSFS_NAME_LEN];
	char type;

	if ((sysfs_get_name_from_path(path, name, SYSFS_NAME_LEN)) != 0)
			return 1;
	if (strncmp(name,"bus",3) == 0)
		type = 'b';
	else if (strncmp(name,"class",5) == 0)
		type = 'c';
	else if (strncmp(name,"devices",7) == 0)
		type = 'r';

	dir = sysfs_open_directory(path);
	if (dir == NULL) {
		fprintf(stderr, "Error opening sysfs directory at %s\n", path);
		return 1;
	}
	if ((sysfs_read_directory(dir)) != 0) {
		fprintf(stderr, "Error reading sysfs directory at %s\n", path);
		return 1;
	}
	current = dir->subdirs;
	memset(name,0,SYSFS_NAME_LEN);
	while (current != NULL) {
		if ((sysfs_get_name_from_path(current->path, name,
		    SYSFS_NAME_LEN)) != 0) {
			sysfs_close_directory(dir);
			return 1;
		}
		switch (type) {
		case 'b':
			memset(BUS_NAME,0,SYSFS_NAME_LEN);
			strncpy(BUS_NAME,name,strlen(name));
			list_bus();
			break;
		case 'c':
			memset(CLASS_NAME,0,SYSFS_NAME_LEN);
			strncpy(CLASS_NAME,name,strlen(name));
			list_class();
			break;
		case 'r':
			memset(TOPO_NAME,0,SYSFS_NAME_LEN);
			strncpy(TOPO_NAME,name,strlen(name));
			show_sysfs_root(TOPO_NAME);
			break;
		default:
			break;
		}	
		current = current->next;
	}
	sysfs_close_directory(dir);
	return 0;
}

/**
 * show_default_info: prints current buses, classes, and root devices
 *	supported by sysfs.
 * returns 0 with success or 1 with error.
 */
int show_default_info(void)
{
	char sysfs_root[SYSFS_PATH_MAX];
	char path_to_print[SYSFS_PATH_MAX];
	int retval = 0;
	/* get sysfs mount point */
	if (sysfs_get_mnt_path(sysfs_root, SYSFS_PATH_MAX) != 0) {
		perror("sysfs_get_mnt_path");
		fprintf(stderr, "Error getting sysfs mount point\n");
		exit(1);
	}
	strcpy(path_to_print, sysfs_root);

	/* print supported buses */
	strcat(path_to_print, SYSFS_BUS_DIR);
	fprintf(stdout, "Supported buses:\n");
	if (DIAG_OPTIONS & DIAG_SHOW_VALUE)
		retval = show_all_device(path_to_print);
	else
		retval = show_subdirectories(path_to_print, 4);

	if (retval == 0) {
		/* print supported classes */
		strcpy(path_to_print, sysfs_root);
		strcat(path_to_print, SYSFS_CLASS_DIR);
		fprintf(stdout, "Supported classes:\n");
		if (DIAG_OPTIONS & DIAG_SHOW_VALUE)
			show_all_device(path_to_print);
		else
			retval = show_subdirectories(path_to_print, 4);
	}
	if (retval == 0) {
		/* print supported root devices */
		strcpy(path_to_print, sysfs_root);
		strcat(path_to_print, SYSFS_DEVICES_DIR);
		fprintf(stdout, "Supported root devices:\n");

		if (DIAG_OPTIONS & DIAG_SHOW_VALUE)
			show_all_device(path_to_print);
		else
			retval = show_subdirectories(path_to_print, 4);
	}
	return retval;
}


int list_main()
{
	
	if ((DIAG_OPTIONS & DIAG_DRIVER) == 0)
		DIAG_OPTIONS |= DIAG_DEVICE;
	if (BUS_NAME[0] != 0)
		list_bus();
	if (CLASS_NAME[0] != 0)
		list_class();
	if (BUS_NAME[0] == 0 && CLASS_NAME[0] == 0 && TOPO_NAME[0] == 0 )
		show_default_info();
	return 0;

}

